# *******************************************************************************
#
#    Copyright (c) 2020-2021 David Briant. All rights reserved.
#
# *******************************************************************************

import sys
if hasattr(sys, '_ImportTrace') and sys._ImportTrace: print(__name__)


from ._sentinels import Missing
import itertools

_idSeed = itertools.count(start=0)
_nameSeed = itertools.count(start=1)



class BType(object):
    # basic type - {id, name}
    # types are essentially global singletons
    #   * __new__ is really a get operation
    #   * ensure is used to ensure a type exists (i.e. creates it if it doesn't exist)
    #   * new creates a new type type and will error if it already exists
    _BTypeByName = {}
    @classmethod
    def new(cls, name):
        # creates a new type throwing an error if it already exists
        if (cls._BTypeByName.get(name, Missing)) is not Missing:
            raise TypeError(f'Type "{name}" already exists')
        instance = super().__new__(BTNominal if cls is BType else cls)
        instance.id = next(_idSeed)
        instance.name = 't' + str(next(_nameSeed)) if name is Missing else name
        instance._factoryFn = Missing
        cls._BTypeByName[instance.name] = instance
        return instance
    @classmethod
    def ensure(cls, name):
        # creates a new type if one does not already exists with the provided name
        if (instance := cls._BTypeByName.get(name, Missing)) is Missing:
            instance = cls.new(name)
        return instance
    def __new__(cls, name):
        # gets a type throwing an error if it has not already been defined
        if (instance := cls._BTypeByName.get(name, Missing)) is Missing:
            raise TypeError(f'Unknown type "{name}"')
        return instance
    def __instancecheck__(self, x):
        if hasattr(x, '_t'):
            return x._t in self
        return type(x) in self

    def __add__(self, other):   # +   - creates a sum type combining self and other
        types = [self] + (other.types if isinstance(other, BTSum) else [other])
        return BTSum(*types)
    def __radd__(self, other):
        return self+other   # commutes so reuse self + other
    def __mul__(self, other):   # *   - creates a tuple type combining self and other
        types = [self] + (other.types if isinstance(other, BTTuple) else [other])
        return BTTuple(*types)
    def __rmul__(self, other):   # *   - creates a tuple type combining self and other
        types = (other.types if isinstance(other, BTTuple) else [other]) + [self]
        return BTTuple(*types)

    # A -> B is B ** A  - describes a finite size map (e.g. an array, hash map etc)
    def __pow__(self, other):
        # other -> self is self ** other
        return BTArray(other, self)
    def __rpow__(self, other):
        # self -> other is other ** self
        return BTArray(self, other)

    # A -> B is B ^ A - describes an infinite size map (e.g. a function)
    def __xor__(self, other):
        # other -> self is self ^ other
        return BTSignature(other, self)
    def __rxor__(self, other):
        # self -> other is other ^ self
        return BTSignature(self, other)

    def __str__(self):
        return self.__repr__()
    # all non-union types are a union of length 1
    def __len__(self):
        return 1
    def __contains__(self, item):
        return item == self
    def __repr__(self):
        return self.name
    def __hash__(self):
        return hash(self.id)
    def __eq__(self, other):
        if not isinstance(other, BType):
            return NotImplemented
        else:
            return self.id == other.id
    def __call__(self, *args, **kwargs):
        if kwargs:
            return BDTT(self, *args, **kwargs)
        elif len(args) == 1:
            return BSTT(self, *args)
        elif len(args) > 1:
            return BMTT(self, *args)
        else:
            raise TypeError('Must provide some tagging args')
    def _killType(self, name):
        del BType._BTypeByName[name]


class BTNominal(BType):
    def __ror__(self, other):  # |  - creates a BAtom if self is a Nominal type else
        return BAtom(self, other) if self._factoryFn is Missing else self._factoryFn(other)


class BTSum(BType):
    # sum type - {id, name, types}
    _BSumTypeByTypes = {}
    def __new__(cls, *args):
        if isinstance(args[0], str):
            name = args[0]
            types = tuple(sorted(args[1:], key=lambda t: t.id if isinstance(t, BType) else hash(t)))
            if not (len(types) > 1): raise TypeError('Must provide at least two types')
            id = types
        else:
            name = Missing
            types = tuple(sorted(args, key=lambda t: t.id if isinstance(t, BType) else hash(t)))
            if not (len(types) > 1): raise TypeError('Must provide at least two types')
            id = types
        if (instance := cls._BSumTypeByTypes.get(id, Missing)) is Missing:
            instance = super().ensure(name)
            instance.types = types
            cls._BSumTypeByTypes[types] = instance
        else:
            if not (name is Missing or name == instance.name):
                raise TypeError(f'Sum type with those types is already named {instance.name}')
        return instance
    def __add__(self, other):
        types = self.types + (other.types if isinstance(other, BTSum) else (other,))
        return BTSum(*types)
    def __len__(self):
        return len(self.types)
    def __contains__(self, item):
        return item in self.types
    def __repr__(self):
        return f"{{{ '+'.join([repr(e) for e in self.types]) }}}"
    def __hash__(self):
        return hash((t.id if isinstance(t, BType) else hash(t) for t in self.types))
    def __eq__(self, other):
        if not isinstance(other, BTSum):
            return False
        else:
            return self.types == other.types


class BTTuple(BType):
    # a heterogenous product (fixed size) type that maps by index - {id, name, types}
    _BTupleTypeByTypes = {}
    def __new__(cls, *args):
        if not args: raise TypeError('Must have at least one type in tuple type')
        if isinstance(args[0], str):
            name = args[0]
            types = args[1:]
            if not types: raise TypeError('Must have at least one type in tuple type')
            id = types
        else:
            name = Missing
            types = args
            # if not types: raise TypeError('Must have at least one type in tuple')   # already tested above
            id = types
        if (instance := cls._BTupleTypeByTypes.get(id, Missing)) is Missing:
            instance = super().ensure(name)
            instance.types = list(types)
            cls._BTupleTypeByTypes[types] = instance
        else:
            if name is not Missing and name != instance.name:
                raise TypeError(f'Tuple type with those types is already named {instance.name}')
        return instance
    def __mul__(self, other):
        types = self.types + (other.types if isinstance(other, BTTuple) else [other])
        return BTTuple(*types)
    def __repr__(self):
        return '(' + ('*'.join([repr(t) for t in self.types])) + ')'
    def __hash__(self):
        return hash((hash(t) for t in self.types))
    def __eq__(self, other):
        if not isinstance(other, BTTuple):
            return False
        else:
            return self.types == other.types


class BTStruct(BType):
    # a heterogenous product (fixed size) type that maps by field - {id, name, typeByField}
    _BTStructByTypes = {}
    def __new__(cls, *args, **kwargs):
        if args:
            if not (len(args) == 1): raise TypeError('Only the typename may be provided as non keyword argument')
            if not (isinstance(args[0], str)): raise TypeError('Name must be a string (if provided)')
            name = args[0]
            fields = tuple(kwargs.keys())
            types = tuple(kwargs.values())
        else:
            name = Missing
            fields = tuple(kwargs.keys())
            types = tuple(kwargs.values())
        if not types:
            return super().ensure(name)
        if (instance := cls._BTStructByTypes.get((fields, types), Missing)) is Missing:
            instance = super().ensure(name)
            instance.typeByField = kwargs
            cls._BTStructByTypes[(fields,types)] = instance
        return instance
    @property
    def fields(self):
        return self.typeByField.keys()


class BTaggedType(BType): pass


class BSTT(BTaggedType):
    # simple tagged type - {id, name, base, parent, [tag=Missing]}
    _BSTTById = {}
    def __new__(cls, parent, *args):
        # args may be name or name, tag or tag (i.e. anonymous)
        if len(args) == 1:
            if isinstance(args[0], str):
                # e.g. num('ccy')       -> id is (matrix, 'ccy')
                name, tag = args[0], Missing
                id = (parent, name)
            elif isinstance(args[0], BType):
                # e.g.  matrix(inout)   -> id is (matrix, 'inout')
                name, tag = Missing, args[0]
                id = parent, (tag.name)
            else:
                raise TypeError('If only one arg provided it must be a name or a type')
        elif len(args) == 2:
            # e.g.  matrix('iomatrix', inout)   -> id is (matrix, 'iomatrix')
            name, tag = args
            id = (parent, name)
        else:
            raise TypeError('Too many args provided')
        if (instance := cls._BSTTById.get(id, Missing)) is Missing:
            instance = super().ensure(name)  # this will fill in a name if it is Missing
            instance.givenName = name is not Missing
            instance.parent = parent
            instance.tag = tag
            instance.base = parent.base if isinstance(parent, BTaggedType) else parent
            cls._BSTTById[(parent, tag)] = instance
        else:
            if not isinstance(instance, BSTT): raise TypeError(f'{parent.name}({id[1]}) is already a {type(instance)}')
            if not (instance.tag == tag): raise TypeError(f'{parent.name}({id[1]}) already exists but with tag {tag}')
        return instance
    def __repr__(self):
        return self.name if self.givenName else (repr(self.parent) + '(' + repr(self.name) + ')')


class BMTT(BTaggedType):
    # multi tagged type - {id, name, base, parent, tags}
    _BMTTById = {}
    def __new__(cls, parent, *args):
        # args must be name, *tags or *tags where tags is 2+
        if not (len(args) >= 2): raise TypeError('Need to provide at least two args')
        if isinstance(args[0], str):
            name = args[0]
            tags = tuple(sorted(args[1:], key=lambda t: t.id))
            if not (len(tags) > 1): raise TypeError('Need to provide at least two tags')
            id = (parent, tags)
        else:
            name = Missing
            tags = tuple(sorted(args, key=lambda t: t.id))
            id = (parent, tags)
        if (instance := cls._BMTTById.get(id, Missing)) is Missing:
            instance = super().ensure(name)        # this will fill in a name if it is Missing
            if hasattr(instance, 'parent'):
                raise TypeError(f'{repr(instance)} already exists')
            instance.parent = parent
            instance.tags = tags
            instance.base = parent.base if isinstance(parent, BTaggedType) else parent
            cls._BMTTById[(parent,tags)] = instance
        else:
            if not isinstance(instance, BMTT): raise TypeError(f'{parent.name}({id[1]}) is already a {type(instance)}')
        if not (instance.tags == tags): raise TypeError(f'{parent.name}({id[1]}) already exists but with tags {tags}')
        if not (name is Missing or instance.name == name):
            raise TypeError(f'Not allowed to rename an exisiting type (i.e. {repr(instance)})')
        return instance
    def __repr__(self):
        return repr(self.parent) + '(' + (','.join([repr(t) for t in self.tags])) + ')'


class BDTT(BTaggedType):
    # detailed tagged type (with key value pairs) - {id, name, base, parent, tags}
    _BDTTById = {}
    def __new__(cls, parent, *nameOrEmpty, **namedtags):
        # signature must be name, **namedtags or *namedtags where namedtags is 1+
        if len(nameOrEmpty) == 1:
            if isinstance(nameOrEmpty[0], str):
                name = nameOrEmpty[0]
                id = (parent, tuple(namedtags.keys()), tuple(namedtags.values()))
            else:
                raise TypeError('Only one keyword argument may be provided and it must be a string')
        else:
            name = Missing
            id = (parent, tuple(namedtags.keys()), tuple(namedtags.values()))
        if (instance := cls._BDTTById.get(id, Missing)) is Missing:
            instance = super().ensure(name)        # this will fill in a name if it is Missing
            name = instance.name
            if hasattr(instance, 'parent'):
                raise TypeError(f'Type with name {repr(name)} already exists - {repr(instance)}')
            instance.parent = parent
            instance.base = parent.base if isinstance(parent, BTaggedType) else parent
            if isinstance(parent, BDTT):
                # any overrides of namedTags must be type compatible i.e. the given type must intersect with the parent type
                for tagname, t in namedtags.items():
                    if tagname in parent.namedtags:
                        if not (_isParentOf(parent.namedtags[tagname], t)):
                            super()._killType(cls, name)
                            raise TypeError(f'{tagname}\'s type is not tagged type of {repr(parent)}')
                for tagname, _ in parent.namedtags.items():
                    if tagname not in namedtags:
                        super()._killType(cls, name)
                        raise TypeError(f'{tagname} from {repr(parent)} must be specified')
            instance.namedtags = namedtags
            cls._BDTTById[(parent, tuple(namedtags.keys()), tuple(namedtags.values()))] = instance
        else:
            if not isinstance(instance, BDTT): raise TypeError(f'{parent.name}({id[1]}) is already a {type(instance)}')
            if not (instance.namedtags == namedtags): raise TypeError(
                f'{parent.name}({id[1]}) already exists but with namedtags {namedtags}')
            if not (instance.parent == parent): raise TypeError(
                f'{parent.name}({id[1]}) already exists but with parent {instance.parent}')
        return instance
    def __repr__(self):
        return repr(self.parent) + f"({'' if self.name[0] == 't' else repr(self.name) + ','}{_dictWithEquals(self.namedtags)})"

def _dictWithEquals(d):
    return ','.join([f'{k}={v}' for k,v in d.items()])

def _isParentOf(parent, x):
    while isinstance(x, BTaggedType):
        if x.parent == parent:
            return True
        x = x.parent
    return False


class BTSignature(BType):
    _BTSignatureById = {}
    # a homogenous and generalised map (exponential) type - aka function
    def __new__(cls, *args):
        # args must be {name, argTypes, retType} or {argTypes, retType}
        if len(args) == 2:
            name = Missing
            argTypes = args[0]
            retType = args[1]
            id = (argTypes, retType)
        elif len(args) == 3:
            name = args[0]
            if not isinstance(name, str): raise TypeError('Name must be a string')
            argTypes = args[1]
            retType = args[2]
            id = (argTypes, retType)
        else:
            raise TypeError('Need to provide either {name, argTypes, retType} or {argTypes, retType}')
        if (instance := cls._BTSignatureById.get(id, Missing)) is Missing:
            instance = super().ensure(name)  # this will fill in a name if it is Missing
            instance.argTypes = argTypes
            instance.retType = retType
            cls._BTSignatureById[id] = instance
        if not (name is Missing or instance.name == name):
            raise TypeError(f'Not allowed to rename an exisiting type (i.e. {repr(instance)})')
        return instance
    def __repr__(self):
        return repr(self.argTypes) + '->' + repr(self.retType)


class BTArray(BType):
    # a homogenous discrete / finite map (exponential) type, where the key is an offset - i.e. N**T, 3**T etc
    _BArrayTypeById = {}
    def __new__(cls, *args):
        # signature must be N, BType
        if not len(args) == 2: raise TypeError(f'Must provide exactly two args: one of {{O, Oa, Ob, Oc}} and a Btype')
        if not args[0] in (O, Oa, Ob, Oc, N, Na, Nb, Nc):
            raise TypeError(f'First arg must be one of {{O, Oa, Ob, Oc}}')
        name = Missing
        id = args
        if (instance := cls._BArrayTypeById.get(id, Missing)) is not Missing:
            pass
        else:
            instance = super().ensure(name)
            instance.indexType = args[0]
            instance.mappedType = args[1]
            cls._BArrayTypeById[id] = instance
        return instance
    def __repr__(self):
        return f'{self.mappedType}**{self.indexType}'


class BTMap(BType):
    # a homogenous discrete / finite map (exponential) type, where the key is not an index or offset - e.g. T**S S->T
    def __init__(self, count, type):
        super().__init__('', '')
        self.index = index
        self.type = type
    def __repr__(self):
        return f'{self.index}**{self.type}'


class BAtom(object):
    # immutable holder of a nominal type and a value
    __slots__ = ['_t_', '_v_']
    def __init__(self, _t, _v):
        assert isinstance(_t, (BType, type))
        self._t_ = _t
        self._v_ = _v
    def __setattr__(self, key, value):
        if key in ('_t_', '_v_'):
            BAtom.__dict__[key].__set__(self, value)
        else:
            raise AttributeError()
    @property
    def _t(self):
        return self._t_
    @property
    def _v(self):
        return self._v_
    def __repr__(self):
        return f'BAtom({self._t_},{self._v_})'
    def __str__(self):
        return f'<{self._t_}:{self._v_}>'
    def __eq__(self, other):
        if not isinstance(other, BAtom):
            return False
        else:
            return (self._t_ == other._t_) and (self._v_ == other._v_)


class Holder(object):
    # a rebindable holder of a sum-type and a tv
    __slots__ = ['_st_', '_tv_']
    def __init__(self, _st, *args):
        self._st_ = _st
        self._tv_ = args[0] if len(args) == 1 else Missing
    def __setattr__(self, key, value):
        if key in ('_st_', '_tv_'):
            Holder.__dict__[key].__set__(self, value)
        elif key == '_tv':
            if not isinstance(value, self._st_): raise TypeError()
            Holder.__dict__['_tv_'].__set__(self, value)
        else:
            raise AttributeError()
    @property
    def _st(self):
        return self._st_
    @property
    def _tv(self):
        return self._tv_
    def __repr__(self):
        return f'``Holder``({self._st_},{self._tv_})'


class BIndex(int):
    def __new__(cls, v, *args, **kwargs):
        return super(cls, cls).__new__(cls, v)
    @property
    def _v(self):
        return self #super().__new__(int, self)
    @property
    def _t(self):
        return tIndex
    def __repr__(self):
        return f'i{super().__repr__()}'

tIndex = BType.new('index')
tIndex._factoryFn = BIndex
N = tIndex('N')
Na = tIndex('Na')
Nb = tIndex('Nb')
Nc = tIndex('Nc')


class BCount(int):
    def __new__(cls, v, *args, **kwargs):
        return super(cls, cls).__new__(cls, v)
    def __add__(self, other):
        return NotImplemented
    def __sub__(self, other):
        return NotImplemented
    def __mul__(self, other):
        return NotImplemented
    def __div__(self, other):
        return NotImplemented
    @property
    def _t(self):
        return tCount
    @property
    def _v(self):
        return self #super().__new__(int, self)
    def __repr__(self):
        return f'c{super().__repr__()}'

tCount = BType.new('count')
tCount._factoryFn = BCount


class BOffset(int):
    def __new__(cls, v, *args, **kwargs):
        return super(cls, cls).__new__(cls, v)
    @property
    def _t(self):
        return tOffset
    def _v(self):
        return self
    def __repr__(self):
        return f'o{super().__repr__()}'

tOffset = BType.new('offset')
tOffset._factoryFn = BOffset
O = tOffset('O')
Oa = tOffset('Oa')
Ob = tOffset('Ob')
Oc = tOffset('Oc')


class BNum(float):
    def __new__(cls, v, *args, **kwargs):
        return super(cls, cls).__new__(cls, v)
    def __add__(self, other):
        res = super(BNum, self).__add__(other)
        return self.__class__(res)
    def __sub__(self, other):
        res = super(BNum, self).__sub__(other)
        return self.__class__(res)
    def __mul__(self, other):
        res = super(BNum, self).__mul__(other)
        return self.__class__(res)
    def __truediv__(self, other):
        res = super(BNum, self).__truediv__(other)
        return self.__class__(res)
    @property
    def _v(self):
        return super().__new__(float, self)
    @property
    def _t(self):
        return tNum
    def __repr__(self):
        return f'n{super().__repr__()}'

tNum = BType.new('num')
tNum._factoryFn = BNum



# templateArg = BType.new('templateArg')
# T = templateArg('T')
# T1 = templateArg('T1')
# T2 = templateArg('T2')
# T3 = templateArg('T3')
# T4 = templateArg('T4')

