# *******************************************************************************
#
#    Copyright (c) 2020-2021 David Briant. All rights reserved.
#
# *******************************************************************************


from coppertop import assertRaises, Missing

from bones.types import BType, Holder
import bones.types


tFred = BType.ensure('fred')
tJoe = BType.ensure('joe')
tSally = BType.ensure('sally')



def testNominalAndAtoms():
    index = BType('index')
    count = BType('count')
    offset = BType('offset')
    num = BType('num')

    assert repr(tFred) == 'fred'
    assert tFred == BType('fred')
    assert tFred is BType('fred')

    x = 'hello' | tFred
    assert isinstance(x, tFred)
    assert x._v == 'hello'

    i1 = 1 | index
    assert i1._t == index
    c1 = 1 | count
    assert c1._t == count
    o1 = 1 | offset
    assert o1._t == offset
    n1 = 1 | num
    assert n1._t == num


def testHolder():
    pystr = BType.ensure('pystr')
    fred = Holder(pystr+int)
    assert fred._st == pystr+int
    assert fred._tv == Missing
    with assertRaises(TypeError):
        fred._tv = 'joe'
    with assertRaises(AttributeError):
        fred._st = str
    fred._tv = ('hello' | pystr)
    assert fred._tv._t == pystr
    fred._tv = 1
    assert fred._tv == 1
    assert isinstance(fred._tv, int)
    assert isinstance(fred._tv, int+pystr)


def testBTSum():
    s = tFred+tJoe

    assert (repr(s) == '{fred+joe}') or (repr(s) == '{joe+fred}')
    assert tFred in s
    assert tSally not in s
    assert tJoe+tFred == s   # summing is commutative

    assert isinstance('hello' | tFred, s)


def testBTTuple():
    p1 = tFred*tJoe
    assert repr(p1) == '(fred*joe)'
    assert p1 != tJoe*tFred              # the product is not commutative
    assert p1 == tFred*tJoe


def testBTSimpleTagged():
    num = BType('num')
    t1 = num('ccy')
    assert t1.parent == num
    assert t1.name == 'ccy'
    assert t1.tag == Missing

    matrix = BType.ensure('matrix')
    inout = BType.ensure('inout')

    t2 = matrix(inout)
    assert t2.parent == matrix
    assert t2.name[0:1] == 't'
    assert t2.tag == inout


def testBTMultiTagged():
    matrix = BType.ensure('matrix')
    inout = BType.ensure('inout')
    xxx = BType.ensure('xxx')
    yyy = BType.ensure('yyy')

    t1 = matrix(inout, xxx)
    assert t1.parent == matrix
    assert t1.name[0:1] == 't'
    assert t1.tags == matrix(xxx, inout).tags

    t2 = matrix('destructableMatrix', inout, yyy)
    assert t2.parent == matrix
    assert t2.name == 'destructableMatrix'
    assert t2.tags == matrix('destructableMatrix', yyy, inout).tags


def testBTDetailedTagged():
    num = BType('num')
    ccy = num('ccy')
    GBP = ccy('GBP')
    USD = ccy('USD')
    GBPUSD1 = num('GBPUSD1', domestic=GBP, foreign=USD)
    assert GBPUSD1.parent == num
    assert GBPUSD1.namedtags['domestic'] == GBP
    assert GBPUSD1.namedtags['foreign'] == USD

    fx = num('fx', domestic=ccy, foreign=ccy)
    assert fx.parent == num
    assert fx.namedtags['domestic'] == ccy
    assert fx.namedtags['foreign'] == ccy

    GBPUSD2 = fx('GBPUSD2', domestic=GBP, foreign=USD, fred=num)
    assert GBPUSD2.parent == fx
    assert GBPUSD2.base == num
    assert GBPUSD2.namedtags['domestic'] == GBP
    assert GBPUSD2.namedtags['foreign'] == USD

    with assertRaises(TypeError) as ar:
        fx('FREDFRED', domestic=GBP)
    assert ar.exceptionValue.args[0] == 'foreign from num(\'fx\',domestic=ccy,foreign=ccy) must be specified'

    with assertRaises(TypeError) as ar:
        fx('FREDFRED', domestic=GBP, foreign=BType('index'))
    assert ar.exceptionValue.args[0] == 'foreign\'s type is not tagged type of num(\'fx\',domestic=ccy,foreign=ccy)'


def testBTArray():
    null = BType.ensure('null')
    num = BType.ensure('num')
    N = BType('N')
    tArr = (num+null) ** N
    assert repr(tArr) == f'{num+null}**N'
    assert isinstance(tArr, bones.types._types.BTArray)


def testBTSignature():
    null = BType.ensure('null')
    fn = (tSally+null) ^ (tFred*tJoe)
    assert repr(fn) == '(fred*joe)->{sally+null}'
    assert isinstance(fn, bones.types._types.BTSignature)


def main():
    testNominalAndAtoms()
    testHolder()
    testBTSum()
    testBTTuple()
    testBTSimpleTagged()
    testBTMultiTagged()
    testBTDetailedTagged()
    testBTArray()
    testBTSignature()


if __name__ == '__main__':
    main()
    print('pass')
